memory.lhs, an implementation of the memory game, using the Gtk2Hs Haskell GUI library

written by Kenneth Hoste, February 2005

needed files: 1.jpg, 2.jpg, 3.jpg, 4.jpg, 5.jpg, 6.jpg, 7.jpg, 8.jpg, 9.jpg, found.jpg,
              back.jpg, memory.glade

provided in gtk2hs-tmr1.(zip|tar.gz)

> import Graphics.UI.Gtk
> import Graphics.UI.Gtk.Glade
> import Char
> import Data.IORef

----------------
| Declarations |
----------------

  represents a game board with a width, height, and a list of buttons

> data Board = Board Int Int [String]

  represent a game state: which button is currently pressed (0 = none is pressed), and the number of couples left to be found

> data State = State (Maybe ToggleButton) (Maybe (Bool,ToggleButton, ToggleButton)) Int

> getImages :: [String]
> getImages = ["1.jpg","2.jpg","3.jpg","4.jpg","5.jpg","6.jpg","7.jpg","8.jpg","9.jpg"]

--------------------
| Start a new game |
--------------------

> buildNewGame :: Int -> IORef State -> IO Board

> buildNewGame n state = do 
>                         writeIORef state (State Nothing Nothing n)
>			  let imagesPart = take n getImages
>			      images     = imagesPart ++ imagesPart
>			  case n of 
>		 	     1 -> return (Board 2 1 images)
>			     2 -> return (Board 2 2 images)
>			     3 -> return (Board 3 2 images)
>			     4 -> return (Board 3 3 images)
>			     5 -> return (Board 4 3 images)
>			     6 -> return (Board 4 3 images)
>			     7 -> return (Board 4 4 images)
>			     8 -> return (Board 4 4 images)
>			     9 -> return (Board 5 4 images)
                            

-------------------------
| Handle button toggles |
-------------------------

> showImageOnButton :: ToggleButton -> String -> IO()

> showImageOnButton button file = do children <- containerGetChildren button
>			 	     containerRemove button (head children)
>				     image <- imageNewFromFile file
>				     containerAdd button image
>				     widgetShowAll button


> handleClick :: IORef State -> String -> ToggleButton -> State -> Bool -> IO()

  special state, don't do anything

> handleClick _ _ _ (State _ _ (-1)) _ = return ()

  game finished

> handleClick _ _ _ (State Nothing (Just (True,p1,p2)) 0) _ = do showImageOnButton p1 "found.jpg"
>							         showImageOnButton p2 "found.jpg"

  button clicked is a card which is 'solved' already

> handleClick _ "found" _ _ _ = return ()


  second button clicked, and toggle is up 

> handleClick ref name button (State (Just lastButton) _ tot) False = do writeIORef ref (State Nothing Nothing tot)
>								         showImageOnButton button "back.jpg"

  first button clicked when game just started

> handleClick ref name button (State Nothing Nothing tot) _ = do writeIORef ref (State (Just button) Nothing tot)
>								 showImageOnButton button name

  first button clicked when last pair didn't match

> handleClick ref name button (State Nothing (Just (False,p1,p2)) tot) _ = do writeIORef ref (State (Just button) Nothing tot)
>									      showImageOnButton button name
>									      showImageOnButton p1 "back.jpg"
>									      showImageOnButton p2 "back.jpg"

  first button clicked when last pair matched

> handleClick ref name button (State Nothing (Just (True,p1,p2)) tot) _ = do writeIORef ref (State (Just button) Nothing tot)
>									     showImageOnButton button name
>									     showImageOnButton p1 "found.jpg"
>									     showImageOnButton p2 "found.jpg"

  second button clicked, and toggle is down 

> handleClick ref name button (State (Just prevBut) _ tot) True = do last <- widgetGetName prevBut
>								     let matched = (name == last)
>								     writeIORef ref (State Nothing Nothing (-1))
>								     toggleButtonSetActive False button
>								     toggleButtonSetActive False prevBut
>								     prevName <- widgetGetName prevBut
>								     showImageOnButton button name
>								     if (matched) then do widgetSetName button "found"
>								                          widgetSetName prevBut "found"
>								       			  writeIORef ref (State Nothing 
>												(Just (matched,button,prevBut)) (tot-1))
>											  if (tot-1 == 0) 
>												then toggleButtonSetActive True button
>												else return ()
>										  else writeIORef ref (State Nothing 
>											     (Just (matched,button,prevBut)) tot)




> buttonToggled :: ToggleButton -> IORef State -> IO()

> buttonToggled button stateRef = 
>  do                               state <- readIORef stateRef
>                                   name <- widgetGetName button
>				    pressed <- toggleButtonGetActive button
>				    handleClick stateRef name button state pressed

-----------------------------
| Creating a new game board |
-----------------------------

> fillRow :: HBox -> [String] -> IORef State -> IO()

> fillRow box [] state = do return ()

> fillRow box (l:ls) state = do button <- toggleButtonNew
>				widgetSetName button $ l
>				containerSetBorderWidth button 2
>				image <- imageNewFromFile "back.jpg"
>				containerAdd button image
>			        onToggled button (buttonToggled button state)
>                               boxPackStartDefaults box button
>                               fillRow box ls state 

> addHBoxToVBox :: Int -> VBox -> [String] -> IORef State -> IO()

> addHBoxToVBox w vBox list state = do hBox <- hBoxNew True w
>				       fillRow hBox list state
>                                      boxPackStartDefaults vBox hBox

> fillBox :: Board -> IORef State -> IO VBox

> fillBox (Board w 1 list) state = do
>                              	        vBox <- vBoxNew True 1 
>					addHBoxToVBox w vBox list state
>                               	return vBox

> fillBox (Board w 2 list) state = do
>                              		vBox <- vBoxNew True 1 
>					addHBoxToVBox w vBox (take w list) state
>					addHBoxToVBox w vBox (drop w list) state
>                               	return vBox

> fillBox (Board w 3 list) state = do
>                              		vBox <- vBoxNew True 1 
>					addHBoxToVBox w vBox (take w list) state
>					addHBoxToVBox w vBox (take w (drop w list)) state
>					addHBoxToVBox w vBox (drop (2*w) list) state
>                               	return vBox

> fillBox (Board w 4 list) state = do
>                              		vBox <- vBoxNew True 1 
>					addHBoxToVBox w vBox (take w list) state
>					addHBoxToVBox w vBox (take w (drop w list)) state
>					addHBoxToVBox w vBox (take w (drop (2*w) list)) state
>					addHBoxToVBox w vBox (drop (3*w) list) state
>                               	return vBox


---------------------------
| Handling closure events |
---------------------------

  returning 'False' results in destroys the window
  (and throws a destroy-event)

> deleteEvent :: Event -> IO Bool
> deleteEvent _ = do return False

  
  what should happen when the window is being closed

> destroyEvent :: IO()
> destroyEvent = do mainQuit

--------------------
| : : : MAIN : : : |
--------------------

> main :: IO()
> main = do 


    initializing GUI

>   initGUI

    loading glade file

>   windowXmlM <- xmlNew "memory.glade"
>   let windowXml = case windowXmlM of
>	 (Just windowXml) -> windowXml
>        Nothing -> error "Can't find the glade file \"memory.glade\" in the current directory"

    loading the window

>   window <- xmlGetWidget windowXml castToWindow "window"
>   onDelete window deleteEvent
>   onDestroy window destroyEvent

    loading control panel

>   controlPanel <- xmlGetWidget windowXml castToVBox "control panel"

    loading entry box

>   entry <- xmlGetWidget windowXml castToEntry "number of pictures"

    loading new game button

>   button <- xmlGetWidget windowXml castToButton "new game button"

    loading message label

>   label <- xmlGetWidget windowXml castToLabel "message label"
>   labelSetText label "\nNothing set."

    loading board alignment

>   boardAlignment <- xmlGetWidget windowXml castToAlignment "board content"

    loading board definition

>   board <- xmlGetWidget windowXml castToVBox "cards"

    preparing game state

>   state <- newIORef (State Nothing Nothing 0)

    setting click actions

>   onClicked button $ do
>      			  entryText <- entryGetText entry
>			  let text = filter (not.isSpace) entryText
>                             check = (not $ null text) && null (tail text) && isDigit (head text)
>                         if (check) then do let level = (read text :: Int)
>					     labelSetText label ("\nNew game started (level: "++(show level)++").") 
>					     cards <- buildNewGame level state
>					     cardsBox <- fillBox cards state
>					     children <- containerGetChildren boardAlignment
>					     if (not $ null children) then containerRemove boardAlignment (head children) else return()
>					     containerAdd boardAlignment cardsBox
>					     widgetShowAll window
>                                    else do labelSetText label ("\nPlease enter a level between\n1 and 9 to set the game level.") 
>					     children <- containerGetChildren boardAlignment
>					     if (not $ null children) then containerRemove boardAlignment (head children) else return()
>					     widgetShowAll window

    packing and showing everything

>   widgetShowAll window

    wait for events to occure

>   mainGUI
